/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client;

import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.pua.prot.client.dataobject.*;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.AtgProtocolRequest;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

/**
 * Klasse zum Abfragen der Ergebnisdatensätze.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolResultStream implements AutoCloseable {

	/** Empfänger der Ergebnisdaten. */
	private final ProtocolAnswerReceiver _receiver;

	/** Statusinformationen. */
	private ProtocolRequestResult _resultInfo;

	/** Zeigt an ob noch mit weiteren Datensätzen zu rechnen ist. */
	private boolean _done;

	/** Zeigt an dass die Protokollerstellung abgebrochen wurde. */
	private boolean _abort;

	/** Objekt zum deserialisieren. */
	private SerializerUtil _serializer = null;

	protected ProtocolAdministrationData _protocolAdministrationData;

	protected ProtocolAggregationData _protocolAggregationData;

	private static final Debug _debug = Debug.getLogger();

	private long _dataNumber = 0;

	private PuaClient _puaClient;

	/**
	 * @param receiver Empfänger der Ergebnisdaten.
	 * @param result   Zeigt ob das Absetzen des Auftrags zur Protokollerstellung erfolgreich war.
	 * @param puaClient PuaClient
	 * @throws FailureException Fehler bei der Initialisierung.
	 */
	public ProtocolResultStream(ProtocolAnswerReceiver receiver, ProtocolRequestResult result, final PuaClient puaClient) throws FailureException {
		_puaClient = puaClient;
		_serializer = new SerializerUtil();
		this._receiver = receiver;
		this._resultInfo = result;
		this._done = false;
		this._abort = false;
	}

	/**
	 * Zeigt den Status der Aktion ab. Der Status kann sich nach jedem Aufruf von {@link #take()} verändern. Wird beispielsweise ein Protokoll erstellt, so ist
	 * wird die Methode <code>true</code> zurückliefern, wenn die Anfrage an den Server gesendet werden konnte. D.h. <code>true</code> bedeutet in diesem Fall
	 * nicht, dass das Protokoll erfolgreich erstellt wurde, sondern nur, dass die Anfrage an den Server gesendet wurde. Ruft man nun {@link #take()} auf, und
	 * anschließend wieder {@link #isRequestSuccessful()}, dann wird z.B. <code>false</code> geliefert, wenn es zu einem fehler bei der Protokollerstellung kam -
	 * etwa durch einen Syntaxfehler im Skript.
	 *
	 * @return True, falls die Aktion momentan keine Fehler meldet.
	 */
	public boolean isRequestSuccessful() {
		return _resultInfo.isRequestSuccessful();
	}

	/**
	 * Zeigt an ob das Protokoll vollständig und erfolgreich übertragen wurde.
	 *
	 * @return <code>true</code> falls das Protokoll vollständig übertragen wurde.<code>false</code> falls noch nicht alle Ergebnisdatensätze eingetroffen sind oder
	 *         das Protokoll abgebrochen wurde.
	 */
	public boolean isComplete() {
		return !wasAborted() && !hasNext() && _resultInfo.isRequestSuccessful();
	}

	/** @return Liefert eine Fehlermeldung zurück, falls es zu einem Fehler kam. */
	public String getErrorMsg() {
		return _resultInfo.getErrorMsg();
	}

	/**
	 * Liefert jeweils einen Antwortdatensatz. Blockiert so lange, bis ein Antwortdatensatz vorliegt. Sobald der letzte Datensatz oder ein Fehler empfangen wurde,
	 * meldet sich das Objekt beim Datenverteiler als Empfänger ab.
	 *
	 * @return Antwortdatensatz. <code>null</code> falls bereits alle Antwortdatensätze ausgeliefert wurden oder ein Fehler aufgetreten ist.
	 *
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public ProtocolResult take() throws InterruptedException {
		ProtocolResult result = null;
		if(_done){
			_resultInfo.set("Verbindung zum Server wurde geschlossen.");
			return null;
		}
		if(_resultInfo.isRequestSuccessful()) {
			final ResultData resultData = _receiver.take(0);
			if(resultData == null) {
				abort();
				if(_protocolAdministrationData == null)
				{
					// Wenn noch keine Daten eingetroffen sind, hängen wir in der Warteschlange fest
					_resultInfo.set("Der Server antwortet nicht. Vermutlich ist die maximale Zahl laufender Auswertungen erreicht.");
				}
				else
				{
					_resultInfo.set("Server hat Protokollübertragung abgebrochen.");
				}
				return null;
			}
			final AtgProtocolRequest answer = AtgProtocolRequest.getJavaObject(resultData.getData());
			if(answer.opCode == ProtocolRelatedOperation.ACK_FLAG) {
				done();
			}
			else if(answer.opCode == ProtocolRelatedOperation.ABORT) {
				aborted(resultData);
			}
			else if(answer.opCode == ProtocolRelatedOperation.NACK_FLAG) {
				done();
				_resultInfo = new ProtocolRequestResult(SerializerUtil.deserialize(answer.requestData).toString());
			}
			else {
				// deserialisieren:
				try {
					final Object obj = _serializer.deserializeZIP(answer.requestData);
					if(answer.opCode == ProtocolRelatedOperation.CREATE && obj instanceof ProtocolResult) {
						result = (ProtocolResult)obj;
						if(result instanceof ProtocolAdministrationData) {
							_protocolAdministrationData = (ProtocolAdministrationData)result;
						}
						else if(result instanceof ProtocolAggregationData) {
							_protocolAggregationData = (ProtocolAggregationData)result;
						}
						else if(result instanceof ProtocolResultData) {
							final ProtocolResultData protocolResultData = (ProtocolResultData)result;
							protocolResultData.setAdminData(_protocolAdministrationData);
						}
						else if(result instanceof ProtocolAggregationResultData) {
							final ProtocolAggregationResultData protocolAggregationResultData = (ProtocolAggregationResultData)result;
							protocolAggregationResultData.setAdminData(_protocolAdministrationData);
							protocolAggregationResultData.setAggregationData(_protocolAggregationData);
						}
						autoFlowControl();
					}
					else {
						_resultInfo = new ProtocolRequestResult(obj.toString());
						result = null;
						_receiver.unsubscribe();
					}
				}
				catch(FailureException e) {
					_debug.warning("Fehler beim Empfang des Protokolls", e);
					_resultInfo = new ProtocolRequestResult(e.getMessage());
					result = null;
					_receiver.unsubscribe();
				}
			}
		}
		return result;
	}

	private void autoFlowControl() {
		final int maxConcurrentDatasets = _puaClient.getFlowControlWindowSize();
		if(maxConcurrentDatasets == 0) {
			// Flusskontrolle deaktiviert
			return;
		}
		Long jobId = getJobId();
		if(jobId == null || _abort){
			return;
		}
		// Datensatzzähler hochzählen
		_dataNumber++;
		if(_dataNumber > 0 && _dataNumber % maxConcurrentDatasets == 0){
			// Pua-Client soll weitere 2 * maxConcurrentDatasets Datensätze an den Client senden
			// damit sind kurz vor dem Versand des nächsten Flusskontrolldatensatzes immer noch etwa maxConcurrentDatasets Datensätze
			// im Transfer, was verhindert dass der Client leer läuft.
			_puaClient.sendFlowControl(jobId, _dataNumber + maxConcurrentDatasets * 2);
		}
	}

	/** Empfang abgeschlossen. Meldet den Receiver ab. */
	private void done() {
		_done = true;
		if(_receiver != null) _receiver.unsubscribe();
		if(_serializer != null) _serializer.done();
		_serializer = null;
	}

	/**
	 * Zeigt an dass keine weiteren Ergebnisdatensätze mehr eintreffen werden. {@link #take()} wird genau 1x <code>null</code> zurückliefern, bevor
	 * <code>isDone</code> <code>true</code> zurückliefert.
	 *
	 * @return <code>true</code> falls alle Ergebnisdatensätze abgeholt wurden.
	 */
	public boolean hasNext() {
		return !_done;
	}

	/**
	 * Meldet das Objekt beim Datenverteiler als Empfänger ab.
	 *
	 * @param data Daten von Pua.
	 */
	private void aborted(ResultData data) {
		String errorMsg = "";
		Object o = SerializerUtil.deserialize(data.getData().getItem(PidProtocol.data).asUnscaledArray().getByteArray());
		if(o != null) {
			if(o instanceof String) {
				errorMsg = (String)o;
			}
			else {
				errorMsg = ErrorMessage.UNEXPECTED_DATA_TYPE + o.getClass().getName() + " <-> String, in ProtocolResultStream.abort().";
			}
		}
		else {
			errorMsg = "Server hat Vorgang abgebrochen";
		}
		aborted(errorMsg);
	}

	/**
	 * Meldet das Objekt beim Datenverteiler als Empfänger ab.
	 *
	 * @param errorMsg Grund des Abbruchs.
	 */
	private void aborted(String errorMsg) {
		_debug.warning("Protokoll wurde abgebrochen", errorMsg);
		_resultInfo.set(errorMsg);
		_abort = true;
		done();
	}

	/** 
	 * Meldet das Objekt beim Datenverteiler als Empfänger ab.
	 * @deprecated Unklarer Name, da die Protokollerstellung nciht abgebrochen wird, sodnern nur der Abruf der Daten. Durch {@link #close()} ersetzt. 
	 */
	@Deprecated
	public void abort() {
		this.close();
	}

	/**
	 * Zeigt ob die Protokollerstellung abgebrochen wurde.
	 *
	 * @return	<code>true</code> falls die Protokollerstellung abgebrochen wurde.
	 */
	public boolean wasAborted() {
		return _abort;
	}

	/**
	 * Liefert die Auftrags-Id.
	 *
	 * @return Auftrags-Id oder <code>null</code> falls die Auftrags-Id unbekannt ist. Bei Protokoll-Erstellen-Aufträgen ist die Auftrags-ID
	 * identisch zur Protokoll-Id.
	 */
	public Long getJobId() {
		if(_receiver == null) return null;
		return _receiver.getJobId();
	}


	/**
	 * Liefert die Auftrags-Id. <b>Wichtig:</b> Wird diese Methode direkt nach dem Absetzen des Auftrages zur Protokollerstellung abgesetzt, ist die Auftrags-Id
	 * noch unbekannt, und es wird <code>null</code> zurückgegeben!
	 *
	 * @return Auftrags-Id oder <code>null</code> falls die Auftrags-Id noch unbekannt ist. Bei Protokoll-Erstellen-Aufträgen ist die Auftrags-ID
	 * identisch zur Protokoll-Id.
	 *
	 * @deprecated Diese Methode liefert nicht mehr die Protokoll-Id, sondern die eindeutige ID des protokollbezogenen Auftrags
	 */
	@Deprecated
	public final Long getProtocolId() {
		return getJobId();
	}

	/**
	 * Liefert die Anzahl der im Client-Puffer liegenden Datensätze.
	 *
	 * @return Anzahl der im Client-Puffer liegenden Datensätze.
	 */
	public int bufferedDataSize() {
		if(_receiver == null) return 0;
		return _receiver.incoming.size();
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		if(_resultInfo.isRequestSuccessful()) {
			return super.toString();
		}
		else {
			return super.toString() + ": " + this.getErrorMsg();
		}
	}

	/**
	 * Beendet die Übertragung der Protokolldaten. Der PuA-Server wird darüber informiert, dass weitere Datensätze nicht übertragen werden müssen.
	 * Sollte das Protokoll gespeichert werden, wird nach diesem Aufruf das Protokoll nur noch gespeichert, aber nicht online übertragen. 
	 */
	@Override
	public void close() {
		if(_done) return;
		Long jobId = getJobId();
		if(jobId != null) {
			_puaClient.abortProtocol(jobId, true);
		}
		aborted("Protokollerstellung abgebrochen");
	}
}
