/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager.taskmanager;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.manager.status.StatePublisher;
import de.bsvrz.pua.prot.processing.Processing;
import de.bsvrz.pua.prot.processing.ProcessingInterface;
import de.bsvrz.pua.prot.processing.ProtocolReader;
import de.bsvrz.pua.prot.sender.PuaSender;
import de.bsvrz.pua.prot.util.JobInProgress;
import de.bsvrz.pua.prot.util.JobState;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.AtlMeta;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

/**
 * Thread der einen Auftrag der in den Berecih der Protokollerstellung fällt ausführt. Mögliche Aufträge sind:<br> - Protokoll erstellen<br> - Protokoll
 * abbrechen<br> - Status der Protokollerstellung anfordern<br>
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ThreadElement extends Thread {

	/** Status der Protokollerstellung nicht ermittelbar. */
	static public final byte INVALID_STATUS = -1;

	public enum Operation {

		NONE(ProtocolRelatedOperation.INVALID),
		CREATE_PROTOCOL(ProtocolRelatedOperation.CREATE),
		SEND_SAVED_PROTOCOL(ProtocolRelatedOperation.SEND_SAVED_PROTOCOL),
		SEND_UNREAD_PROTOCOL(ProtocolRelatedOperation.SEND_UNREAD_PROTOCOL);

		private final int _code;

		Operation(final int code) {
			_code = code;
		}

		public int getCode() {
			return _code;
		}

	}
	/** 1: Protokollerstellung. 2: Gespeichertes Protokoll senden. 3: Noch nicht abgerufenes Protokoll senden. */
	private Operation _operation = Operation.NONE;

	/** Auftragsparameter */
	private ProcessingParameter _pp = null;

	/** Intitale ziel-Datensatznummer der Flusskontrolle */
	private long _initialFlowControl;

	/** Objekt das die Datenaufbereitung durchführt */
	private ProcessingInterface _processor = null;

	/** Erzeuger diser ThreadElement Instanz */
	private ThreadPool _creator = null;

	/** True falls der Thread die Datenaufbereitung beenden soll. */
	private boolean _abort = false;

	/** Falls der Thread vollständig beendet werden soll. */
	private volatile boolean _die = false;

	/** Debug-Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/** Zeigt an, ob der Thread bereits läuft, d.h. ob er sich bereits in die <code>run</code> Methode befindet. */
	public boolean running = false;

	/** Data Manager. Ermöglicht das Interpretieren des Skripts */
	private final DataManagerInterface dataManager;

	/** Protokoll-Id */
	private long _protocolId;

	/** Auftrags_ID */
	private long _jobId;

	/** Id der initialen Anfrage zur Protokollerstellung. */
	private long _requestId;

	/** Objekt zum Senden der Daten */
	private PuaSender _sender;

	/** Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann. */
	private final long _maxDelay;

	/**
	 * Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird
	 */
	private final int _timeout;

	/** true falls ein Auftrag vorliegt und noch abgearbeitet werden muss. */
	private boolean _busy;

	private boolean _inProgress = false;
	private boolean _hasError = false;

	/**
	 * Zeitpunkt an dem der Auftrag erstellt wurde
	 */
	private long _startTime;

	/** Objekt zum serialisieren und zippen der Ergebnisdaten. */
	private SerializerUtil _serializer = null;

	/**
	 * Initialisierung des Objekts
	 *
	 * @param creator     Erzeuger des Threads. Wird benachrichtigt (@link ThreadPool#finished(ThreadElement)}), sobald Thread unbeschäftigt ist.
	 * @param dataManager Data Manager. Ermöglicht das Interpretieren des Skripts
	 * @param maxDelay    Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann.
	 *
	 * @param timeout Timeout
	 * @throws FailureException Fehler bei der Initialisierung.
	 */
	public ThreadElement(ThreadPool creator, DataManagerInterface dataManager, long maxDelay, final int timeout) throws FailureException {
		_timeout = timeout;
		_serializer = new SerializerUtil();
		this._creator = creator;
		this.dataManager = dataManager;
		this._maxDelay = maxDelay;
	}

	/**
	 * Liefert den Status des Auftrags
	 *
	 * @return Prozentualer Fertigstellungsgrad des Auftrags oder {@link #INVALID_STATUS} falls kein Auftrag anliegt.
	 */
	public byte getStatus() {
		if(_processor != null) {
			return _processor.getStatus();
		}
		else {
			return INVALID_STATUS;
		}
	}

	/**
	 * Evaluiert das Skript. Dazu wird über das Submodul Daten Manager ({@link DataManagerInterface}) der Skriptinterpreter ({@link
	 * de.bsvrz.pua.prot.interpreter.Interpreter}) verwendet.
	 *
	 * @return Informationen für die Datenaufbereitung
	 *
	 * @throws FailureException Fehler im Skript.
	 */
	protected ProcessingInformation evaluateScript() throws FailureException {
		return dataManager.process(_pp);
	}

	/**
	 * Bricht die Protokollerstellung ab. Der Thread meldet dem {@link #_creator}, dass er wieder für neue Anfragen zur Verfügung steht. Dazu ruft er die Methode
	 * {@link ThreadPool#finished(ThreadElement)} auf. Wird das Protokoll gespeichert auf dem Datenträger gespeichert, wird es wieder gelöscht. Im Gegensatz zu
	 * {@link #kill} steht der Thread danach weiterhin zur Verfügung.
	 */
	public void abort() {
		_debug.finer("Thread " + this.getName() + ": Protokollerstellung wird abgebrochen");
		_abort = true;
		if(_processor != null) _processor.abort();

		/*
				 * Wichtig: Hier nicht creator.finished() aufrufen!
				 * Dies wird NUR in der run() Methode getan,
				 * sonst kann das Objekt nicht mehr gelockt werden! (für synchronized)
				 */
	}

	public void stopProtocolTransmission() {
		if(_processor != null) _processor.stopProtocolTransmission();
	}
	
	/** Startet den Thread und versetzt ihn sofort in einen Schlafzustand, aus dem er erst wieder geweckt wird, wenn eine Anfrage eingeht. */
	@Override
    public void run() {
		try {
			while(!_die) {
				try {
					_debug.finer("Thread " + this.getName() + " beendet Anfragenbearbeitung & wartet");
					running = true;
					//falls Thread nach Fertigstellung eines Protokolls sofort wieder
					//einen Auftrag zugewiesen wird ist es möglich, dass activate(...)
					//VOR erreichen von wait(...)aufgerufen wird -> es muss geprüft
					//werden, ob überhaupt gewartet werden soll.
					synchronized(this){
						while(!_busy)
						{
							this.wait();
						}
					}
					if(!_die && _busy) {
						_inProgress = true;
						_hasError = false;
						_debug.finer("Thread " + this.getName() + " beginnt mit Anfragenbearbeitung");
						try {
							switch(_operation) {
								case CREATE_PROTOCOL:		//Protokollerstellung
									//Anfrage abarbeiten
									ProcessingInformation pi = evaluateScript();
									processData(pi);
									break;
								case SEND_SAVED_PROTOCOL: 
								case SEND_UNREAD_PROTOCOL: // Fallthrough
									sendProtocol();
									break;
								default:
									_hasError = true;
									_debug.warning("Ungültige Operation: " + _operation);
							}
						}
						catch(FailureException fe) {
							sendError(
									new FailureException(
											"Fehler bei der Protokollerstellung durch Thread " + this.getName() + ", " + fe.getMessage(), fe.getLogLevel()
									)
							);
							_hasError = true;
						}
					}
				}
				catch(InterruptedException e) {    //exception wird nicht gesetzt, da es sich nicht um einen Fehler handelt (InterruptedException wird geworfen, wenn die DA abgebrochen werden soll).
					_debug.finest(ErrorMessage.INTERRUPTED_EXCEPTION + _protocolId);
				}
				catch(Throwable e){
					_debug.error("Unbehandelter Fehler bei der Bearbeitung einer Anfrage", e);
					sendError(
							new FailureException(
									"Fehler bei der Protokollerstellung durch Thread " + this.getName() + ", " + e.getMessage(), LoggerException.ERROR
							)
					);
					_hasError = true;
				}
				interrupted(); // Interrupted-Status löschen
				_inProgress = false;

				if(_busy && _operation != Operation.NONE) {
					assert _sender.getDav() != null;
					StatePublisher.getInstance(_sender.getDav()).notifyThreadFinished(this);
				}

				//Abbrechen nicht nötig
				_abort = false;
				_processor = null;
				//Erledigt melden & sender abmelden
				if(_sender != null) {
					_sender = null;
				}
				this.setName("IdleThread (was " + this.getName() + ")");
				_busy = false;                //Wichtig! finshed ruft u.U. wieder activate... auf. Dort wird busy wieder auf true gesetzt! Deshalb MUSS buys=false VOR creator.finished stehen!
				_creator.finished(this);
			}
			_debug.finer("Thread " + this.getName() + " hat die Anfragenbearbeitung beendet.");

		}
		finally {
			if(_sender != null) {
				_sender = null;
			}
			this.setName("IdleThread (was " + this.getName() + ")");
			_serializer.done();
			_serializer = null;
		}
	}

	/**
	 * Überträgt ein Protokoll an den Auftraggeber. Liest die einzelnen Datensätze aus der Datei ein, und legt sie auf den Datenverteiler.
	 *
	 * @throws FailureException
	 */
	private void sendProtocol() {
		_processor = new ProtocolReader(_sender, _serializer, _protocolId, _maxDelay, _timeout, dataManager, _jobId, _initialFlowControl);
		_processor.process(_abort);
	}

	/**
	 * Sendet eine Fehlermeldung an den Auftraggeber und gibt eine Debug-Nachricht aus.
	 *
	 * @param exception Fehlermeldung.
	 */
	private void sendError(FailureException exception) {
		exception.log();
		try {
			_debug.finest("Setze Fehlermeldung ab: " + _requestId);
			ProtocolAnswer.sendError(_sender, _requestId, exception.getMessage());
		}
		catch(FailureException e) {
			_debug.error(ErrorMessage.CAN_NOT_SEND + e.getMessage());
		}
	}

	/**
	 * Führt die Datenaufbereitung durch.
	 *
	 * @return Ergebnis der Datenaufbereitung.
	 *
	 * @throws FailureException Metainformation des Skripts ist korrupt.
	 */
	private void processData(ProcessingInformation pi) throws FailureException {
		AtlMeta meta = dataManager.getMetaInformation(_pp.getScript());
		if(meta == null) {
			String msg = ManagerErrorMessage.NO_META_INFO + _pp.getScript();
			sendError(new FailureException(msg, LoggerException.WARNING));
		}
		else {
			_processor = new Processing(_sender, dataManager, _serializer, _protocolId, _pp, pi, meta, _creator.getConfigAuth(), _maxDelay, _timeout, _initialFlowControl);
			_processor.process(_abort);
		}
	}

	/**
	 * Startet die Protokollerstellung. Setzt die nötigen Parameter für die Datenaufbereitung und startet die Datenaufbereitung.
	 *
	 * @param dav        Verbindung zum Datenverteiler
	 * @param client     Auftraggeber
	 * @param requestId  Anfrage-ID
	 * @param pp         Auftragsparameter.
	 * @param protocolId Protokoll-Id
	 *
	 * @param jobId JobId
	 * @param startTime StartTime
	 * @param initialFlowControl InitialFlowControl
	 * @throws FailureException Fehler bei der Sendeanmeldung.
	 */
	public void activateCreateProtocol(ClientDavInterface dav, SystemObject client,
			long requestId, ProcessingParameter pp, long protocolId, final long jobId, final long startTime, final long initialFlowControl)
			throws FailureException {
		this._sender = getPuaSender(dav, client);
		// Thread-Steuerung:
		this._busy = true;
		this._operation = Operation.CREATE_PROTOCOL;
		//Auftragsparameter übernehmen.
		this.setName("CreateProtocol Pid:" + protocolId );
		this._protocolId = protocolId;
		this._requestId = requestId;
		this._jobId = jobId;
		this._startTime = startTime;
		this._pp = pp;
		this._initialFlowControl = initialFlowControl;
		synchronized(this) {
			notifyAll();
		}
	}

	/**
	 * Startet das Senden eines bereits gespeicherten Protokolls.
	 *
	 * @param dav        Verbindung zum Datenverteiler.
	 * @param client     Auftraggeber.
	 * @param requestId  Anfrage-ID
	 * @param protocolId Id des Protokolls, das gesendet werden soll.
	 *
	 * @param jobId JobId
	 * @param startTime StartTime
	 * @param initialFlowControl InitialFlowControl
	 * @throws FailureException Fehler bei der Sendeanmeldung.
	 */
	public void activateGetProtocol(ClientDavInterface dav, SystemObject client, long requestId, long protocolId, final long jobId, final long startTime, final long initialFlowControl) throws FailureException {
		this._sender = getPuaSender(dav, client);
		// Thread-Steuerung:
		this._busy = true;
		this._operation = Operation.SEND_SAVED_PROTOCOL;
		//Auftragsparameter übernehmen.
		this.setName("SendSavedProtocol Pid:"  + protocolId + " JobId:" + jobId);
		this._protocolId = protocolId;
		this._requestId = requestId;
		this._jobId = jobId;
		this._startTime = startTime;
		this._pp = null;
		this._initialFlowControl = initialFlowControl;
		synchronized(this) {
			notifyAll();
		}
	}

	/**
	 * Startet das Senden eines noch nicht abgerufenen Protokolls.
	 *
	 * @param dav        Verbindung zum Datenverteiler.
	 * @param client     Auftraggeber.
	 * @param requestId  Anfrage-ID
	 * @param protocolId Id des Protokolls, das gesendet werden soll.
	 *
	 * @param jobId JobId
	 * @param startTime StartTime
	 * @param initialFlowControl InitialFlowControl
	 * @throws FailureException Fehler bei der Sendeanmeldung.
	 */
	public void activateGetUnreadProtocol(ClientDavInterface dav, SystemObject client, long requestId, long protocolId, final long jobId, final long startTime, final long initialFlowControl) throws FailureException {
		this._sender = getPuaSender(dav, client);
		// Thread-Steuerung:
		this._busy = true;
		this._operation = Operation.SEND_UNREAD_PROTOCOL;
		//Auftragsparameter übernehmen.
		this.setName("SendUnreadProtocol Pid:"  + protocolId + " JobId:" + jobId);
		this._protocolId = protocolId;
		this._requestId = requestId;
		this._jobId = jobId;
		this._startTime = startTime;
		this._pp = null;
		this._initialFlowControl = initialFlowControl;
		synchronized(this) {
			notifyAll();
		}
	}

	private PuaSender getPuaSender(final ClientDavInterface dav, final SystemObject client) {
		return new PuaSender(dav, client, new DataDescription(dav.getDataModel().getAttributeGroup(PidProtocol.atgProtocolRequest), 
		                                                      dav.getDataModel().getAspect(PidProtocol.aspAnswer)),
		                     () -> {});
	}

	/** Bricht den Auftrag ab und beendet den Thread vollständig. */
	public void kill() {
		_die = true;
		abort();
		synchronized(this) {
			notifyAll();
		}
	}

	/**
	 * Liefert die Protokoll-Id.
	 *
	 * @return Id des Protokolls, das erstellt wird.
	 */
	public long getProtocolId() {
		return _protocolId;
	}

	/**
	 * Gibt die eindeutige ID dieses Auftrags zurück. Diese Id wird statt der Protokoll-ID verwendet, da je Protokoll u.U. mehrere
	 * Aufträge gleichzeitig ausgeführt werden können.
	 * @return ID des Auftrags.
	 */
	public long getJobId() {
		return _jobId;
	}

	/** Flußsteuerung: Versand des Protokolls anhalten. */
	public void pauseProcessing() {
		if(_processor != null) _processor.pauseProcessing();
	}

	/** Flußsteuerung: Versand des Protokolls fortsetzen. */
	public void resumeProcessing() {
		if(_processor != null) _processor.resumeProcessing();
	}


	public void flowControl(final long numData) {
		if(_processor != null) _processor.flowControl(numData);
	}


	/** @return Liefert die initiale Anfrage-Id. */
	public long getRequestId() {
		return _requestId;
	}

	/**
	 * Gibt die aktuell durchgeführte Aktion zurück
	 * @return 1: Protokollerstellung. 2: Gespeichertes Protokoll senden. 3: Noch nicht abgerufenes Protokoll senden.*/
	public Operation getOperation() {
		return _operation;
	}

	/**
	 * Erstellt ein JobInProgress-Objekt zur Übertragung des Auftragsstatus an den Client
	 *
	 * @return JobInprogress
	 * @throws de.bsvrz.sys.funclib.losb.exceptions.FailureException Exception
	 */
	public JobInProgress toJobInProgress() throws FailureException {
		JobState state = JobState.InProgress;
		if(_processor != null){
			state = _processor.getJobState();
		}
		PuaSender sender = _sender;
		if(sender == null){
			throw new FailureException("Client-ID nicht ermittelbar", LoggerException.ERROR);
		}
		return new JobInProgress(_jobId, _operation.getCode(), state, getStatus(), _startTime, sender.getObject().getId(), _pp != null ? _pp : _protocolId);
	}

}
