/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing.archivebuffer;

import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.pua.prot.aggregations.Aggregation;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.IntermediateDataSet;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProcessingInformation.ApplyAggregation;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.pua.prot.util.ResultValue;
import de.bsvrz.pua.prot.util.attributes.AttributeIdentifier;
import de.bsvrz.pua.prot.util.attributes.DurationAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.Elements;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Führt das Aggregieren von Daten durch. Spalten-Aggregation: Ein Aggregationsergebnis für alle Attribute, die sich nur im verwendeten Objekt unterscheiden.
 * D.h. mehreren Attributen wird das gleiche Aggregationsobjekt zugeordnet. Objekt-Aggregation: Ein Aggregationsergebnis pro Attribut. Jedem einzelnen Attribut
 * wird ein eigenes Aggregationsobjekt zugeordnet.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Aggregator {

	/** Aggregationen für Intervall Spalte. In dieser Liste kann eine Aggregationsliste mehrfach vorkommen. */
	private final List<List<Aggregation>> _aggrIntervalAttribute;

	/** Aggregationen für Intervall Objekt. In dieser Liste kann eine Aggregationsliste nur einmal vorkommen. */
	private final List<List<Aggregation>> _aggrIntervalObject;

	/** Aggregationen für Gesamt Spalte */
	private final List<List<Aggregation>> _aggrAllAttribute;

	/** Aggregationen für Gesamt Objekt */
	private final List<List<Aggregation>> _aggrAllObject;

	/** Anwendung "Liste" */
	private final boolean _list;

	/** Zuletzt erhaltener Ausgangsdatensatz */
	private List<BaseDataSet> _mostRecent;

	/** Zuletzt erstellte Ausgangsdatensatz für Interval-Spalte-Aggregation. (Wird für "verchmelzen" und Änderungs / Zustandsprotokoll benötigt. */
	private List<BaseDataSet> _mostRecentIntervalAttributeData;

	/** Zuletzt erstellte Ausgangsdatensatz für Interval-Objekt-Aggregation. (Wird für "verchmelzen" und Änderungs / Zustandsprotokoll benötigt. */
	private List<BaseDataSet> _mostRecentIntervalObjectData;

	/** Änderungsprotokoll oder Zustandsprotokoll. */
	private final ProtocolType _protocolType;

	/** Anzahl der Spalten */
	private final int _numColumns;

	/**
	 * Zeigt an welche Attribute durch die Aggregation <code>spalte</code> zusammengefasst werden. Zusammengehörige Spalten werden duch die gleichen Nummern
	 * gekennzeichnet.
	 */
	private int[] _linkedAttributes = null;

	private final Map<List<List<Aggregation>>, Interval> _intervals;

	private final class Interval{
		private long _start;
		private long _end;

		private Interval(final long start, final long end) {
			_start = start;
			_end = end;
		}

		public long getStart() {
			return _start;
		}

		public long getEnd() {
			return _end;
		}

		public void setStart(final long start) {
			_start = start;
		}

		public void setEnd(final long end) {
			_end = end;
		}

		@Override
		public String toString() {
			return "Interval{" + new Date(_start) + "-" + new Date(_end) + '}';
		}
	}

	/**
	 * @param pi Informationen für die Datenaufbereitung.
	 *
	 * @throws FailureException Fehler beim Anlegen von neuen Instanzen.
	 */
	public Aggregator(final ProcessingInformation pi) throws FailureException {
		_mostRecent = null;
		_mostRecentIntervalAttributeData = _mostRecentIntervalObjectData = null;
		_protocolType = pi.getProtocolType();
		_numColumns = pi.getAttributes().size();

		final List<ProcessingInformation.ApplyAggregation> order = pi.getOrder();

		_intervals=new HashMap<List<List<Aggregation>>, Interval>(4);

		if(order.isEmpty()) {
			_list = true;
			_aggrAllObject = null;
			_aggrAllAttribute = null;
			_aggrIntervalObject = null;
			_aggrIntervalAttribute = null;
		}
		else {
			if(order.contains(ProcessingInformation.ApplyAggregation.GESAMT_OBJEKT)) {
				_aggrAllObject = createAggregationForEveryAttribute(pi.getAttributes());
			}
			else {
				_aggrAllObject = null;
			}

			if(order.contains(ProcessingInformation.ApplyAggregation.GESAMT_SPALTE)) {
				_aggrAllAttribute = createAggregationPerColumn(pi.getAttributes());
			}
			else {
				_aggrAllAttribute = null;
			}

			if(order.contains(ProcessingInformation.ApplyAggregation.INTERVALL_OBJEKT)) {
				_aggrIntervalObject = createAggregationForEveryAttribute(pi.getAttributes());
			}
			else {
				_aggrIntervalObject = null;
			}

			if(order.contains(ProcessingInformation.ApplyAggregation.INTERVALL_SPALTE)) {
				_aggrIntervalAttribute = createAggregationPerColumn(pi.getAttributes());
			}
			else {
				_aggrIntervalAttribute = null;
			}

			if(order.contains(ProcessingInformation.ApplyAggregation.LISTE)) {
				_list = true;
			}
			else {
				_list = false;
			}
		}
	}

	/**
	 * Ordnet jedem Attribut eine Aggregation zu. Attributen, die sich nur durch das Objekt unterscheiden, wird das gleiche Aggregationsobjekt zugeordnet.
	 * Temporären Attributen wird keine Aggregation zugeordnet.
	 *
	 * @param elements Attribute
	 *
	 * @return Neues Feld mit neu instanziierten Objekten.
	 *
	 * @throws FailureException Fehler beim Erzeugen von neuen Instanzen
	 */
	private List<List<Aggregation>> createAggregationPerColumn(final List<Elements> elements) throws FailureException {
		/**
		 * Ermöglicht die Zuordnung Atg:Asp:Name -> Aggregation
		 *
		 * @author beck et al. projects GmbH
		 * @author Martin Hilgers
		 */
		class Key {

			private final AttributeGroup atg;

			private final String asp;

			private final AttributeIdentifier attribute;

			/** @param rad Attributbeschreibung */
			public Key(final RealAttributeDescription rad) {
				atg = rad.qualifier.attributeGroup;
				asp = rad.qualifier.getAspect();
				attribute = rad.getAttributeIdentifier();
			}

			/** @see java.lang.Object#equals(java.lang.Object) */
			@Override
			public boolean equals(final Object obj) {
				if(obj instanceof Key) {
					final Key key = (Key)obj;
					return atg.equals(key.atg) && asp.equals(key.asp) && attribute.equals(key.attribute);
				}
				else {
					return false;
				}
			}

			/** @see java.lang.Object#hashCode() */
			@Override
			public int hashCode() {
				return (37 * ((37 * atg.hashCode() + asp.hashCode()))) + attribute.hashCode();
			}
		}

		if(_linkedAttributes == null) {
			_linkedAttributes = new int[elements.size()];
			for(int i = 0; i < elements.size(); i++) {
				_linkedAttributes[i] = -1;
			}
		}
		int c = 1;
		final HashMap<Key, Tuple<Integer, List<Aggregation>>> sos = new HashMap<Key, Tuple<Integer, List<Aggregation>>>();
		final List<List<Aggregation>> result = new ArrayList<List<Aggregation>>(elements.size());
		for(int i = 0, elementsSize = elements.size(); i < elementsSize; i++) {
			final Elements element = elements.get(i);
			if(element.getAggregationClasses().size() > 0) {
				if(element instanceof RealAttributeDescription) {
					final RealAttributeDescription rad = (RealAttributeDescription)element;
					final Key key = new Key(rad);
					final Tuple<Integer, List<Aggregation>> res = sos.get(key);
					if(res != null) {
						result.add(res.last);
						_linkedAttributes[i] = res.first;
					}
					else {
						final List<Aggregation> aggregationList = createAggregationList(element);
						sos.put(key, new Tuple<Integer, List<Aggregation>>(c, aggregationList));
						result.add(aggregationList);
						_linkedAttributes[i] = c;
						c++;
					}
				}
				// Falls es mehrere Zeitdauer-Attribute gibt, erhält jedes eine eigene Aggregation.
				else if(element instanceof DurationAttributeDescription) {
					final DurationAttributeDescription dad = (DurationAttributeDescription)element;
					result.add(createAggregationList(dad));
					// c wird nicht inkrementiert, da die "Zeitdauern"-Spalten ja mit keiner anderen Spalte
					// zusammengefasst werden -> c anzupassen ist nicht nötig.
				}
				// Temporäre Attribute erhalten keine Aggregation. Attributgruppen können per Definition nicht
				// aggregiert werden.
				else {
					result.add(Collections.<Aggregation>emptyList());
				}
			}
			else {
				result.add(Collections.<Aggregation>emptyList());
			}
		}
		return result;
	}

	/**
	 * Legt ein Aggregationsobjekt pro Systemobjekt an.
	 *
	 * @param elements Datenidentifikationen
	 *
	 * @return Zweifach verschachtelte Liste. Enthält die Aggregationen pro Attribut bzw. pro Spalte.
	 *
	 * @throws FailureException Fehler beim Instanziieren von Aggregationsobjekten.
	 */
	private List<List<Aggregation>> createAggregationForEveryAttribute(final List<Elements> elements) throws FailureException {
		final List<List<Aggregation>> result = new ArrayList<List<Aggregation>>(elements.size());
		for(final Elements element : elements) {
			result.add(createAggregationList(element));
		}
		return result;
	}

	private List<Aggregation> createAggregationList(final Elements element) throws FailureException {
		final List<Class<? extends Aggregation>> aggregationClasses = element.getAggregationClasses();
		final List<Aggregation> result = new ArrayList<Aggregation>(aggregationClasses.size());
		for(final Class<? extends Aggregation> aggregationClass : aggregationClasses) {
			try {
				result.add(aggregationClass.newInstance());
			}
			catch(InstantiationException e) {
				throw new FailureException(InterpreterErrorMessage.CANT_CREATE_AGGREGATIONOBJECT + e.getMessage(), e, LoggerException.WARNING);
			}
			catch(IllegalAccessException e) {
				throw new FailureException(InterpreterErrorMessage.CANT_CREATE_AGGREGATIONOBJECT + e.getMessage(), e, LoggerException.WARNING);
			}
		}
		return result;
	}

	/**
	 * Führt die Aggregation durch. Wird für jeden Ausgangsdatensatz einmal aufgreufen.
	 *
	 * @param bds Liste
	 *
	 * @return Falls der Ausgangsdatensatz an den Auftraggeber weitergeleitet werden soll. Dies ist der Fall, wenn die Anwendung "liste" verwendet wurde.
	 */
	public boolean aggregate(final List<BaseDataSet> bds) {
		_mostRecent = bds;
		for(int i = 0; i < bds.size(); i++) {
			final BaseDataSet baseDataSet = bds.get(i);
			aggregateList(i, baseDataSet, _aggrAllAttribute);
			aggregateList(i, baseDataSet, _aggrAllObject);
			aggregateList(i, baseDataSet, _aggrIntervalAttribute);
			aggregateList(i, baseDataSet, _aggrIntervalObject);
		}
		return _list;
	}

	private void aggregateList(final int i, final BaseDataSet dataSet, final List<List<Aggregation>> attribute) {

		if(attribute == null) return;

		// Intervall für diese Aggregation aktualisieren
		if(dataSet.getDataTimeStamp() > 0) {
			
			final Interval interval = _intervals.get(attribute);
			if(interval == null) {
				// Neues Intervall für diesen Datensatz anlegen
				_intervals.put(attribute, new Interval(dataSet.getDataTimeStamp(), dataSet.getDataTimeStamp()));
			}
			else {
				// Vorhandenes Intervall ausweiten
				interval.setStart(Math.min(interval.getStart(), dataSet.getDataTimeStamp()));
				interval.setEnd(Math.max(interval.getEnd(), dataSet.getDataTimeStamp()));
			}
		}
		final List<Aggregation> aggregationList = attribute.get(i);
		if(aggregationList != null) aggregate(aggregationList, dataSet.getValue());
	}

	private void aggregate(final List<Aggregation> aggregations, final ResultValue value) {
		for(final Aggregation aggregation : aggregations) {
			aggregation.aggregate(value);
		}
	}

	/**
	 * Aggregationsdatensätze abrufen, die zum Ende eines Intervalls vorliegen. Muss aufgerufen werden, bevor die Aggregation für ein weiteres Intervall beginnt.
	 *
	 * @param status Status der dem Ausgangsadtensatz zugewiesen wird.
	 *
	 * @param currentPeriod Aktueller Zeitbereich
	 * @return Feld mit Ergebnis der Aggregation für Intervall_Objekt in <code>0</code>, und dem Ergebnis der Aggregation für Intervall_Spalte in <code>1</code>.
	 *         Die Einträge können <code>null</code> sein, falls keine Aggregationsdaten vorliegen.
	 */
	public IntermediateDataSet[] endOfInterval(final byte status, final int currentPeriod) {
		final IntermediateDataSet[] result = {null, null};
		if(_mostRecent != null) {
			if(_aggrIntervalAttribute != null) result[1] = aggregateInterval(status, currentPeriod, ApplyAggregation.INTERVALL_SPALTE);
			if(_aggrIntervalObject != null) result[0] = aggregateInterval(status, currentPeriod, ApplyAggregation.INTERVALL_OBJEKT);
		}
		return result;
	}

	/**
	 * Erstellt einen neuen Ausgangsdatensatz mit den Aggregationsdaten. Zudem wird überprüft, ob der so erzeugte Ausgangsdatensatz mit dem zuletzt gesendeten
	 * übereinstimmt.
	 *
	 * @param status Status
	 *
	 * @param currentPeriod Aktueller Zeitbereich
	 *@param order  Aggregationsanwendung  @return Ausgangsdatensatz oder <code>null</code> falls sich im Vergleich zum letzten Ausgangsdatensatz nicht geändert hat, und ein Deltaprotokoll erstellt
	 *         werden soll.
	 */
	private IntermediateDataSet aggregateInterval(final byte status, final int currentPeriod, final ApplyAggregation order) {
		final List<List<Aggregation>> aggregations;
		if(order == ApplyAggregation.INTERVALL_SPALTE) {
			aggregations = _aggrIntervalAttribute;
		}
		else {
			aggregations = _aggrIntervalObject;
		}
		final List<BaseDataSet> result = new ArrayList<BaseDataSet>(_numColumns);
		// die Aggregationen merken, deren Daten schon abgefragt wurden
		final HashMap<List<Aggregation>, List<ResultValue>> collected = new HashMap<List<Aggregation>, List<ResultValue>>();
		for(int i = 0; i < aggregations.size(); i++) {
			final List<Aggregation> aggregationList = aggregations.get(i);
			List<ResultValue> rv = collected.get(aggregationList);
			if(rv == null) {
				rv = new ArrayList<ResultValue>();
				for(final Aggregation aggregation : aggregationList) {
					addAggregationResult(rv, aggregation);
				}
				collected.put(aggregationList, rv);
			}

			final BaseDataSet e = createBaseDataSet(aggregations, i, new ResultValue(rv));
			result.add(e);
		}
		if(!BaseDataSet.compareBaseData(
				(order == ApplyAggregation.INTERVALL_SPALTE) ? _mostRecentIntervalAttributeData : _mostRecentIntervalObjectData, result
		)) {
			if(order == ApplyAggregation.INTERVALL_SPALTE) {
				_mostRecentIntervalAttributeData = result;
			}
			else {
				_mostRecentIntervalObjectData = result; // Werte sind unterschiedlich -> Ausgangsdatensatz unverändert
			}
			// zurückgeben
			return new IntermediateDataSet(result, status, order, currentPeriod);
		}
		else if(_protocolType == ProtocolType.StatusProtocol) // Werte sind gleich & Zustandsprotokoll -> "Quittung" zurückgeben
		{
			return new IntermediateDataSet(Collections.<BaseDataSet>emptyList(), IntermediateDataSet.NO_CHANGE, order, currentPeriod);
		}
		else if(_protocolType == ProtocolType.DeltaProtocol) {
			// Werte sind gleich, nichts zurückgeben -> nichts wird gesendet.
			return null;
		}
		else if(_protocolType == ProtocolType.EventProtocol) {
			// Aggregationsdatensatz immer zurückgeben
			return new IntermediateDataSet(result, status, order, currentPeriod);
		}
		throw new UnsupportedOperationException("Unbekannte Protokollart: " + _protocolType);
	}

	private BaseDataSet createBaseDataSet(final List<List<Aggregation>> aggregations, final int i, final ResultValue rv) {
		Interval interval = _intervals.get(aggregations);
		if(interval == null) interval = new Interval(BaseDataSet.INVALID_VALUE, BaseDataSet.INVALID_VALUE);
		final BaseDataSet e = new BaseDataSet(
				rv,
				_mostRecent.get(i).getDataState(),
				_mostRecent.get(i).getDataInformation(),
				_mostRecent.get(i).getDataTimeStamp(),
				_mostRecent.get(i).getArchiveTimeStamp(),
				interval.getStart(),
				interval.getEnd(),
				_mostRecent.get(i).getDataIndex(),
				false
		);
		_intervals.remove(aggregations);
		return e;
	}

	/**
	 * Aggregationsdatensätze abrufen, die zum Ende der Protokollerstellung vorliegen.
	 *
	 * @return Feld mit dem Ergebnis der Aggregation für Gesamt_Objekt in <code>[0]</code> und denm Ergebnis der Aggregation für Gesamt_Spalte in <code>[1]</code>.
	 *         Einträge können <code>null</code> sein, falls keine Aggregationsdaten vorliegen.
	 * @param currentPeriod Aktueller Zeitbereich
	 */
	public IntermediateDataSet[] endOfProtocol(final int currentPeriod) {
		if(_mostRecent != null) {
			final IntermediateDataSet[] result = {null, null};
			// die Aggregationen merken, deren Daten schon abgefragt wurden
			final HashMap<List<Aggregation>, List<ResultValue>> collected = new HashMap<List<Aggregation>, List<ResultValue>>();

			if(_aggrAllObject != null) {
				final ArrayList<BaseDataSet> dataSets = new ArrayList<BaseDataSet>(_aggrAllObject.size());
				for(int i = 0; i < _aggrAllObject.size(); i++) {
					if(_aggrAllObject.get(i) != null) {
						final List<ResultValue> rv = new ArrayList<ResultValue>();
						for(final Aggregation aggregation : _aggrAllObject.get(i)) {
							addAggregationResult(rv, aggregation);
						}
						final BaseDataSet e = createBaseDataSet(_aggrAllObject, i, new ResultValue(rv));
						dataSets.add(e);
					}
					else {
						final BaseDataSet e = createBaseDataSet(_aggrAllObject, i, new ResultValue(new ExpressionResult("-")));
						dataSets.add(e);
					}
				}
				result[0] = new IntermediateDataSet(dataSets, (byte)100, ApplyAggregation.GESAMT_OBJEKT, currentPeriod);
			}

			if(_aggrAllAttribute != null) {
				final ArrayList<BaseDataSet> dataSets = new ArrayList<BaseDataSet>(_aggrAllAttribute.size());
				for(int i = 0; i < _aggrAllAttribute.size(); i++) {
					final List<Aggregation> aggregationList = _aggrAllAttribute.get(i);
					if(aggregationList != null) {
						List<ResultValue> rv = collected.get(aggregationList);
						if(rv == null) {
							rv = new ArrayList<ResultValue>();
							for(final Aggregation aggregation : aggregationList) {
								addAggregationResult(rv, aggregation);
							}
							collected.put(aggregationList, rv);
						}
						final BaseDataSet e = createBaseDataSet(_aggrAllAttribute, i, new ResultValue(rv));
						dataSets.add(e);
					}
					else {
						final BaseDataSet e = createBaseDataSet(_aggrAllAttribute, i, new ResultValue(new ExpressionResult("-")));
						dataSets.add(e);
					}
				}
				result[1] = new IntermediateDataSet(dataSets, (byte)100, ApplyAggregation.GESAMT_SPALTE, currentPeriod);
			}
			return result;
		}
		else {
			return new IntermediateDataSet[]{null, null};
		}
	}

	/**
	 * Fügt das Ergebnis einer Aggregation zu einer Liste hinzu
	 * @param rv Liste
	 * @param aggregation Aggregation
	 */
	private void addAggregationResult(final List<ResultValue> rv, final Aggregation aggregation) {
		final ResultValue value = aggregation.getResult();
		if(value != null)
			rv.add(value);
		else
			rv.add(new ResultValue(new ExpressionResult()));
	}

	/**
	 * Werden auch nicht aggregierte Daten versendet?
	 *
	 * @return <code>true</code> falls <code>liste</code> eine der ausgewählten die Aggregationsanwendungen ist.
	 */
	public boolean isListAggregation() {
		return _list;
	}

	/**
	 * Zeigt an welche Attribute durch die Aggregation <code>spalte</code> zusammengefasst werden. Zusammengehörige Spalten werden duch die gleichen Nummern
	 * gekennzeichnet.
	 *
	 * @return Zusammen gehörende Spalten. <code>null</code> falls die <code>spalten</code> Aggregation nicht verwendet wird.
	 */
	public int[] getLinkedAttributes() {
		return _linkedAttributes;
	}
}
