/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing.util;

import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.pua.prot.processing.onlinebuffer.ExtResultData;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.Serializable;

/**
 * Objekt das das Markierungsattribut eines Datensatzes enthält. Siehe TAnf S. 32. Die Methode {@link #createDataInformation(ExtResultData,boolean)} erzeugt
 * eine neue Markierungsinformation. Eine Markierungsinformation kann nur aus einer anderen Markierungsinformation erzeugt werden, deshalb ist {@link
 * #createDataInformation(ExtResultData,boolean)} nicht <code>static</code>.<br> Wichtig: Beim Erzeugen einer Markierungsinformation kann die ursprüngliche
 * Markierung verändert werden! Eine Markierungsinformation kann daher erst als vollständig betrachtet werden, nachdem eine weitere Markierungsinformation aus
 * ihr generiert wurde.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DataInformation implements Serializable {

	private static final long serialVersionUID = 7745143397180551956L;

	/** Zeitrücksprung */
	public boolean rebound = false;

	/** Potentielle Datenlücke (grob) */
	public byte dataGap = 0;

	/** Index springt */
	public static final byte INDEX_NOT_CONTINUOUS = 2;

	/** Neuinitialisierung */
	public static final byte INITIALIZATION = 4;

	/** Potentielle Datenlücke (fein) */
	public static final byte POTENTIAL_DATA_GAP = 8;

	/** Am Ende */
	public static final byte ENDING = 16;

	/** Keine Nutzdaten */
	public byte noPayload = 0;

	/** Keine Quelle */
	public static final byte NO_SOURCE = 2;

	/** Keine Daten */
	public static final byte NO_DATA = 4;

	/** Keine Rechte */
	public static final byte NO_RIGHTS = 8;

	/** Archivinfo */
	public byte archiveInfo = 0;

	/** Daten gelöscht */
	public static final byte ERASED = 2;

	/** Daten ausgelagert */
	public static final byte UNAVAILABLE_BLOCK = 4;

	/** Spezielle Datenart */
	public byte specialDataType = 0;

	/** Nachgeliefert */
	public static final byte DELAYED = 2;

	/** Nachgefordert */
	public static final byte REQUESTED = 4;

	/** Kennzeichnung */
	public byte tag = 0;

	/** Ende (Teil-)Zeitintervall */
	public static final byte END_OF_INTERVAL = 2;

	//Variablen die für die Berechnung verwendet werden:

	/** Bisher größter Datenzeitstempel. */
	private transient long _oldTimeStamp = Long.MIN_VALUE;

	/** Bisher größter Datenindex */
	private transient long _oldDataIndex = Long.MIN_VALUE;

	/** Gibt an ob ein letzter Datensatz noch zu markieren ist */
	private transient boolean _tagLastItem = false;

	/** Gibt an ob es sich um den ersten Aufruf handelt und die Variablen {@link #_oldDataIndex} und {@link #_oldTimeStamp} noch ihre default Werte haben. */
	private transient  boolean _first = true;

	/** True: Es gibt einen Sprung im Datensatzindex, deshalb muss ein leerer Datensatz eingefügt werden! */
	transient boolean _insertEmpty = false;

	/** Zustands oder Deltaprotokoll? */
	transient ProtocolType _protocolType = ProtocolType.Undefined;

	/**
	 * Erzeugt eine Markierungsinformation.
	 *
	 * @param protocolType Markierungsinformation für ein Zustands oder DeltaProtokoll.
	 */
	public DataInformation(ProtocolType protocolType) {
		this._protocolType = protocolType;
	}

	/**
	 * Erzeugt eine Markierungsinformation.
	 *
	 * @param deltaProtocol Markierungsinformation für ein Zustands oder DeltaProtokoll.
	 *
	 * @deprecated Kostruktor mit ProtocolType-Argument erlaubt Definition von beliebiger Protokollart
	 */
	@Deprecated
	public DataInformation(boolean deltaProtocol) {
		this(deltaProtocol ? ProtocolType.DeltaProtocol : ProtocolType.StatusProtocol);
	}

	/**
	 * Erstellt eine Markierungsinformation. Dabei wird sowohl der übergebene Archivdatensatz als auch die Markierungsinformation <code> this</code>
	 * berücksichtigt. Siehe TAnf. Kapitel 5.1.2.4.3. Diese Methode muss auch für den letzten leeren Datensatz aufgerufen werden, der vom Archivsystem geliefert
	 * wird, um die Flags der "Ende-Markierung" korrekt zu setzen.<br> Wichtig: Beim Erzeugen einer Markierungsinformation kann die ursprüngliche Markierung
	 * verändert werden! D.h. dass u.U. das Objekt, dessen <code>createDataInformation</code>-Methode aufgerufen wurde geändert wird!
	 *
	 * @param data       Neu erhaltener Archivdatensatz. Darf <code>null</code> sein. In diesem Fall wird eine neue Markierungsinformation zurückgegeben.
	 * @param timeSorted <code>true</code> falls die Daten nach Datenzeitstempel sortiert eintreffen.
	 *
	 * @return Erstellte Markierungsinformation
	 */
	public DataInformation createDataInformation(ExtResultData data, boolean timeSorted) {
		DataInformation di = new DataInformation(_protocolType);
		if(data == null) return di;

		di._first = false;

		if(_first) {
			di._oldTimeStamp = data.getDataTime();
			di._oldDataIndex = data.getDataIndex();
		}
		else {

			//MARKIERUNG ZEITRÜCKSPRUNG
			if(timeSorted == true)			//nach DZS geordnet
			{
				//timestamp eintragen
				di._oldTimeStamp = data.getDataTime();

				if(Util.didxNoSubscrTime(_oldDataIndex) > Util.didxNoSubscrTime(data.getDataIndex())) {
					if(_protocolType != ProtocolType.DeltaProtocol) {
						rebound = true;
						di.rebound = true;
					}
					di._tagLastItem = true;
					di._oldDataIndex = _oldDataIndex;
				}
				else {
					di._oldDataIndex = data.getDataIndex();
					if(_tagLastItem == true && _protocolType != ProtocolType.DeltaProtocol)	//Den ersten Datensatz der kein Zeitrückspurng ist auch noch markieren.
					{
						di.rebound = true;
					}
				}
			}
			else							//nach DI geordnet
			{
				//Datenindex eintrage
				di._oldDataIndex = data.getDataIndex();

				if(_oldTimeStamp > data.getDataTime()) {
					if(_protocolType != ProtocolType.DeltaProtocol) {
						rebound = true;
						di.rebound = true;
					}
					di._tagLastItem = true;
					di._oldTimeStamp = _oldTimeStamp;
				}
				else {
					di._oldTimeStamp = data.getDataTime();
					if(_tagLastItem == true && _protocolType != ProtocolType.DeltaProtocol) {
						di.rebound = true;
					}
				}
				if(_protocolType != ProtocolType.DeltaProtocol) {
					// Bei Deltaprotokoll keine Datenindizes prüfen, da sonst identische (archivseitg weggelassen) Daten
					// als Datenlücke erkannt werden

					//MARKIERUNG POT DATENLÜCKE FÜR NACH DI-geordnete Spalten
					if(Util.didxIndexSkipped(_oldDataIndex, data.getDataIndex())) {
						dataGap |= POTENTIAL_DATA_GAP | INDEX_NOT_CONTINUOUS;
						//jetzt muss ein leerer Datensatz eingefügt werden.
						di._insertEmpty = true;
					}
				}
			}
//			Sprung im Initialisierungszeitpunkt überprüfen: Dazu die höherwertigen 4 Bytes des Datenindex überprüfen
			if((_oldDataIndex | 0xFFFFFFFFL) != (data.getDataIndex() | 0xFFFFFFFFL)) dataGap |= POTENTIAL_DATA_GAP | INITIALIZATION;

			//Ende DS hat gleichen DZS wie voriger
			if(data.getDataType().equals(DataState.END_OF_ARCHIVE) && _oldTimeStamp == data.getDataTime()) dataGap |= POTENTIAL_DATA_GAP | ENDING;
		}

		//Archivdaten sind als pot Datenlücke markiert
		if(data.getDataType().equals(DataState.POSSIBLE_GAP)) dataGap |= POTENTIAL_DATA_GAP;

		//MARKIERUNG KEINE NUTZDATEN
		if(data.getDataType().equals(DataState.NO_SOURCE)) {
			noPayload |= NO_SOURCE;
		}
		else if(data.getDataType().equals(DataState.NO_RIGHTS)) {
			noPayload |= NO_RIGHTS;
		}
		else if(data.getDataType().equals(DataState.NO_DATA)) {
			noPayload |= NO_DATA;
		}

		//MARKIERUNG ARCHIVINFO
		if(data.getDataType().equals(DataState.DELETED_BLOCK)) {
			archiveInfo = ERASED;
		}
		else if(data.getDataType().equals(DataState.UNAVAILABLE_BLOCK)) archiveInfo = UNAVAILABLE_BLOCK;

		//MARKIERUNG DATENART
		if(data.getDataKind().equals(ArchiveDataKind.ONLINE_DELAYED) || data.getDataKind().equals(ArchiveDataKind.REQUESTED_DELAYED)) {
			di.specialDataType |= DELAYED;
		}
		if(data.getDataKind().equals(ArchiveDataKind.REQUESTED) || data.getDataKind().equals(ArchiveDataKind.REQUESTED_DELAYED)) {
			di.specialDataType |= REQUESTED;
		}

		//MARKIERUNG KENNZEICHNUNG
		if(data.getDataType().equals(DataState.END_OF_ARCHIVE)) tag |= END_OF_INTERVAL;

		return di;
	}

	/**
	 * Zeigt ob ein leerer Datensatz eingefügt werden muss. Dies ist notwendig, wenn der Datenindex springt und die Daten nach Datenindex sortiert sind.
	 *
	 * @return True falls ein leerer Datensatz aufgrund eines Sprungs im Datenindex eingefügt werden soll.
	 */
	public boolean insertEmpty() {
		return _insertEmpty;
	}

	/**
	 * Zeigt ob die Daten zum Auffüllen verwendet werden können.
	 *
	 * @return True wenn die Daten zum Auffüllen verwendet werden können, false sonst.
	 */
	public boolean dataForPadding() {
		if(_protocolType == ProtocolType.EventProtocol) return false;
		return (dataGap == 0 || (noPayload == DataInformation.NO_DATA) || (noPayload == DataInformation.NO_SOURCE) || (archiveInfo != 0)
		        || (specialDataType != 0) || (tag != 0));
	}

	/**
	 * Klont die Markierungsinformation.
	 *
	 * @return Geklonte Markierungsinformation.
	 */
	@Override
    public DataInformation clone() {
		DataInformation di = new DataInformation(_protocolType);
		di.archiveInfo = archiveInfo;
		di.dataGap = dataGap;
		di.noPayload = noPayload;
		di.rebound = rebound;
		di.specialDataType = specialDataType;
		di.tag = tag;
		di._protocolType = _protocolType;

		di._oldDataIndex = _oldDataIndex;
		di._oldTimeStamp = _oldTimeStamp;
		return di;
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof DataInformation) {
			DataInformation other = (DataInformation)object;
			return this.archiveInfo == other.archiveInfo && this.dataGap == other.dataGap && this._first == other._first && this._insertEmpty == other._insertEmpty
			       && this.noPayload == other.noPayload && this._oldDataIndex == other._oldDataIndex && this._oldTimeStamp == other._oldTimeStamp
			       && this.rebound == other.rebound && this._tagLastItem == other._tagLastItem && this.specialDataType == other.specialDataType
			       && this.tag == other.tag;
		}
		else {
			return false;
		}
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		StringBuilder sb = new StringBuilder();
		if(archiveInfo != 0) sb.append(" Archivinfo");
		if(dataGap != 0) sb.append(" Datenlücke");
		if(_insertEmpty) sb.append(" leeren einfügen");
		if(noPayload != 0) sb.append(" keine Daten");
		if(rebound) sb.append(" Zeitrücksprung");
		if(specialDataType != 0) sb.append(" spezielle Datenart");
		if(tag != 0) sb.append(" Gekennzeichnet");
		if(_tagLastItem) sb.append(" Letzter Datensatz ist zu markieren");
		return sb.toString();
	}

	/**
	 * Vergleicht die Markierungsinformation. Im Gegensatz zu {@link #equals(Object)} werden nur die öffentlichen (public) Felder der Markierungsinformation
	 * verglichen.
	 *
	 * @param other Markierungsinformation.
	 *
	 * @return	<code>true</code> falls die beiden Markierungsinformationen übereinstimmen.
	 */
	public boolean cmpInfo(DataInformation other) {
		if(other == null) {
			return false;
		}
		else {
			return this.rebound == other.rebound && this.dataGap == other.dataGap && this.noPayload == other.noPayload && this.archiveInfo == other.archiveInfo
			       && this.specialDataType == other.specialDataType && this.tag == other.tag && this._insertEmpty == other.insertEmpty();
		}
	}
}
