/*
 * Copyright 2015 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.pua.prot.aggregations.AggregationResultValue;
import de.bsvrz.pua.prot.functions.Abs;
import de.bsvrz.pua.prot.functions.ExpressionInterface;
import de.bsvrz.pua.prot.processing.util.ExpressionResultAndState;

import java.util.Arrays;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;

/**
 * Hilfsklasse, die Fehlermeldungen erzeugt.
 *
 * @author Kappich Systemberatung
 */
public class ErrorMessageBuilder {

	public static final EnumSet<ExpressionResult.ResultType> NUMERIC = EnumSet.of(ExpressionResult.ResultType.LONG, ExpressionResult.ResultType.DOUBLE, ExpressionResult.ResultType.BOOL);

	public static String illegalOperation(final ExpressionTree.Operation operation, final ExpressionTree parameter, final ExpressionResultAndState parameterR) {
		if(parameterR.getType() == ExpressionResult.ResultType.ERROR) {
			return parameterR.getResult().toString();
		}
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Ungültiger Datentyp für Ausdruck ");
		if(parameter != null) {
			stringBuilder.append('"').append(operation.getSymbol()).append(parameter).append("\" ");
		}
		else {
			stringBuilder.append('"').append(operation.getSymbol()).append(parameterR.getType()).append("\" ");
		}
		stringBuilder.append("bei Berechnung von: ");
		stringBuilder.append(operation.getSymbol()).append(parameterR.getResult());
		return stringBuilder.toString();
	}

	public static String illegalOperation(final ExpressionTree.Operation operation, final ExpressionTree left, final ExpressionTree right, final ExpressionResultAndState leftR, final ExpressionResultAndState rightR) {
		if(rightR.getType() == ExpressionResult.ResultType.ERROR && leftR.getType() == ExpressionResult.ResultType.ERROR) {
			if(rightR.getResult().toString().equals(leftR.getResult().toString())) {
				return rightR.getResult().toString();
			}
		}
		else if(rightR.getType() == ExpressionResult.ResultType.ERROR) {
			return rightR.getResult().toString();
		}
		else if(leftR.getType() == ExpressionResult.ResultType.ERROR) {
			return leftR.getResult().toString();
		}
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Ungültiger Datentyp für Ausdruck ");
		if(left != null && right != null) {
			stringBuilder.append('"').append(left).append(operation.getSymbol()).append(right).append("\" ");
		}
		else {
			stringBuilder.append('"').append(leftR.getType()).append(operation.getSymbol()).append(rightR.getType()).append("\" ");
		}
		stringBuilder.append("bei Berechnung von: ");
		stringBuilder.append(leftR.getResult()).append(operation.getSymbol()).append(rightR.getResult());
		return stringBuilder.toString();
	}

	public static String illegalArgument(final ExpressionInterface expressionInterface, final List<ExpressionTree> expressionList, final int argumentIndex, final ExpressionResultAndState argumentValue, final EnumSet<ExpressionResult.ResultType> expectedArgumentType) {
		if(argumentValue.getResult().getType() == ExpressionResult.ResultType.ERROR){
			return argumentValue.getResult().getString();
		}
		String functionName = expressionInterface.getFunctionName();
		return "Ungültiger " + (argumentIndex + 1) + ". Parameter beim Aufruf von " + functionName + "(" + formatActualParams(expressionList) + ")" + ". Erwarteter Typ: " + formatExpectedParam(expectedArgumentType) + ", war aber \"" + argumentValue + "\" (" + argumentValue
				.getType() + ")";
	}

	public static String illegalOperation(final ExpressionTree.Operation operation, final ExpressionTree parameter, final ExpressionResult parameterR) {
		if(parameterR.getType() == ExpressionResult.ResultType.ERROR) {
			return parameterR.toString();
		}
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Ungültiger Datentyp für Ausdruck ");
		if(parameter != null) {
			stringBuilder.append('"').append(operation.getSymbol()).append(parameter).append("\" ");
		}
		else {
			stringBuilder.append('"').append(operation.getSymbol()).append(parameterR.getType()).append("\" ");
		}
		stringBuilder.append("bei Berechnung von: ");
		stringBuilder.append(operation.getSymbol()).append(parameterR);
		return stringBuilder.toString();
	}

	public static String illegalOperation(final ExpressionTree.Operation operation, final ExpressionTree left, final ExpressionTree right, final ExpressionResult leftR, final ExpressionResult rightR) {
		if(rightR.getType() == ExpressionResult.ResultType.ERROR && leftR.getType() == ExpressionResult.ResultType.ERROR) {
			if(rightR.toString().equals(leftR.toString())) {
				return rightR.toString();
			}
		}
		else if(rightR.getType() == ExpressionResult.ResultType.ERROR) {
			return rightR.toString();
		}
		else if(leftR.getType() == ExpressionResult.ResultType.ERROR) {
			return leftR.toString();
		}
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Ungültiger Datentyp für Ausdruck ");
		if(left != null && right != null) {
			stringBuilder.append('"').append(left).append(operation.getSymbol()).append(right).append("\" ");
		}
		else {
			stringBuilder.append('"').append(leftR.getType()).append(operation.getSymbol()).append(rightR.getType()).append("\" ");
		}
		stringBuilder.append("bei Berechnung von: ");
		stringBuilder.append(leftR).append(operation.getSymbol()).append(rightR);
		return stringBuilder.toString();
	}

	public static String illegalAggregation(final AggregationResultValue left, final ResultValue right, final String aggregationName) {
		ExpressionResult leftR = left.getValue();
		ExpressionResult rightR = right.getValue();
		if(leftR == null) leftR = new ExpressionResult();
		if(rightR == null) rightR = new ExpressionResult();
		if(rightR.getType() == ExpressionResult.ResultType.ERROR && leftR.getType() == ExpressionResult.ResultType.ERROR) {
			if(rightR.toString().equals(leftR.toString())) {
				return rightR.toString();
			}
		}
		else if(rightR.getType() == ExpressionResult.ResultType.ERROR) {
			return rightR.toString();
		}
		else if(leftR.getType() == ExpressionResult.ResultType.ERROR) {
			return leftR.toString();
		}
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Ungültige Berechnung bei Aggregation ").append(aggregationName).append(" mit ");
		stringBuilder.append("Zwischenergebnis: \"").append(left).append("\" und neuem Wert: \"").append(right).append("\"");
		stringBuilder.append(" bei Index ").append(left.getAggregationCount() + 1);
		return stringBuilder.toString();
	}

	public static String divisionByZero(final ExpressionTree.Operation operation, final ExpressionTree left, final ExpressionTree right, final ExpressionResultAndState leftR, final ExpressionResultAndState rightR) {
		if(rightR.getType() == ExpressionResult.ResultType.ERROR && leftR.getType() == ExpressionResult.ResultType.ERROR) {
			if(rightR.getResult().toString().equals(leftR.getResult().toString())) {
				return rightR.getResult().toString();
			}
		}
		else if(rightR.getType() == ExpressionResult.ResultType.ERROR) {
			return rightR.getResult().toString();
		}
		else if(leftR.getType() == ExpressionResult.ResultType.ERROR) {
			return leftR.getResult().toString();
		}
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Division durch 0 an Ausdruck ");
		if(left != null && right != null) {
			stringBuilder.append('"').append(left).append(operation.getSymbol()).append(right).append("\" ");
		}
		else {
			stringBuilder.append('"').append(leftR.getType()).append(operation.getSymbol()).append(rightR.getType()).append("\" ");
		}
		stringBuilder.append("bei Berechnung von: ");
		stringBuilder.append(leftR.getResult()).append(operation.getSymbol()).append(rightR.getResult());
		return stringBuilder.toString();
	}

	public static String divisionByZero(final ExpressionTree.Operation operation, final ExpressionTree left, final ExpressionTree right, final ExpressionResult leftR, final ExpressionResult rightR) {
		if(rightR.getType() == ExpressionResult.ResultType.ERROR && leftR.getType() == ExpressionResult.ResultType.ERROR) {
			if(rightR.toString().equals(leftR.toString())) {
				return rightR.toString();
			}
		}
		else if(rightR.getType() == ExpressionResult.ResultType.ERROR) {
			return rightR.toString();
		}
		else if(leftR.getType() == ExpressionResult.ResultType.ERROR) {
			return leftR.toString();
		}
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Division durch 0 an Ausdruck ");
		if(left != null && right != null) {
			stringBuilder.append('"').append(left).append(operation.getSymbol()).append(right).append("\" ");
		}
		else {
			stringBuilder.append('"').append(leftR.getType()).append(operation.getSymbol()).append(rightR.getType()).append("\" ");
		}
		stringBuilder.append("bei Berechnung von: ");
		stringBuilder.append(leftR).append(operation.getSymbol()).append(rightR);
		return stringBuilder.toString();
	}

	@SafeVarargs
	public static String illegalArgumentCount(final ExpressionInterface expressionInterface, final List<ExpressionTree> expressionList, final boolean varargs, final EnumSet<ExpressionResult.ResultType>... expectedArgumentTypes) {
		String functionName = expressionInterface.getFunctionName();
		return "Ungültige Anzahl Parameter beim Aufruf von " + functionName + "(" + formatActualParams(expressionList) + ")" + ". Erwartete Parameter: " + functionName + "(" + formatExpectedParams(expectedArgumentTypes, varargs) + ")";
	}

	private static String formatActualParams(final List<ExpressionTree> expressionList) {
		if(expressionList == null) return "";
		StringBuilder stringBuilder = new StringBuilder();
		for(Iterator<ExpressionTree> iterator = expressionList.iterator(); iterator.hasNext(); ) {
			final ExpressionTree expressionTree = iterator.next();
			stringBuilder.append(expressionTree);
			if(iterator.hasNext()) {
				stringBuilder.append(", ");
			}
		}
		return stringBuilder.toString();
	}

	private static String formatExpectedParams(final EnumSet<ExpressionResult.ResultType>[] resultTypes, boolean varargs) {
		StringBuilder stringBuilder = new StringBuilder();
		for(Iterator<EnumSet<ExpressionResult.ResultType>> iterator = Arrays.asList(resultTypes).iterator(); iterator.hasNext(); ) {
			final EnumSet<ExpressionResult.ResultType> param = iterator.next();
			if(param.equals(NUMERIC)){
				stringBuilder.append("NUMBER");
			}
			else {
				for(Iterator<ExpressionResult.ResultType> inner = param.iterator(); inner.hasNext(); ) {
					final ExpressionResult.ResultType resultType = inner.next();
					stringBuilder.append(resultType);
					if(inner.hasNext()) {
						stringBuilder.append('|');
					}
				}
			}
			if(iterator.hasNext()) {
				stringBuilder.append(", ");
			}
			else if(varargs) {
				stringBuilder.append("...");
			}
		}
		return stringBuilder.toString();
	}

	private static String formatExpectedParam(final EnumSet<ExpressionResult.ResultType> param) {
		StringBuilder stringBuilder = new StringBuilder();
		if(param.equals(NUMERIC)){
			stringBuilder.append("NUMBER");
		}
		else {
			for(Iterator<ExpressionResult.ResultType> inner = param.iterator(); inner.hasNext(); ) {
				final ExpressionResult.ResultType resultType = inner.next();
				stringBuilder.append(resultType);
				if(inner.hasNext()) {
					stringBuilder.append('|');
				}
			}
		}
		return stringBuilder.toString();
	}
}
